/*

   A C++ program that illustrates the following
   small class hierarchy that uses multiple
   inheritance with a common base class:


            ALCalc01
               |
               |
               |
         /------------\
         |            |
         |            |
         |            |
         |            |
         |            |
      ALCalc100     ALCalc200
         |            |
         |            |
         |            |
         \------------/
                |
                |
                |
             ALCalc120

*/

#include <iostream.h>
#include <math.h>
#include <string.h>

const double BAD_RESULT = -1.0e+30;
const double EPSILON = 1.0e-50;

enum Logical { FALSE, TRUE };

class ALCalc01
{
  public:
    ALCalc01(double fRegX = 0., double fRegY = 0.);

    // set new register values
    void setRegX(double fRegX)
      { m_fRegX = fRegX; }
    void setRegY(double fRegY)
      { m_fRegY = fRegY; }
    double getRegZ()
      { return m_fRegZ; }
    Logical getError();

    void add(Logical bShowResult = TRUE,
             const char* pszMsg = "");
    void sub(Logical bShowResult = TRUE,
             const char* pszMsg = "");
    void mult(Logical bShowResult = TRUE,
              const char* pszMsg = "");
    void div(Logical bShowResult = TRUE,
             const char* pszMsg = "");
    void showRegZ(const char* pszMsg = "");
    void showOperation(const char* pszMsg = "");
    void clearRegs();

  protected:
    double m_fRegX;
    double m_fRegY;
    double m_fRegZ;
    char m_cOp;
    Logical m_bErr;

};

class ALCalc100 : virtual public ALCalc01
{
  public:
    ALCalc100(int nMaxMem = 10,
              double fRegX = 0.,
              double fRegY = 0.);
    ~ALCalc100();
    void stoMem(int nIndex, double x);
    void rclMem(int nIndex, double& x);
    void clearMem();

  protected:
    double* m_pfMem;
    int m_nMaxMem;
};

class ALCalc200 : virtual public ALCalc01
{
  public:
    ALCalc200(double fRegX = 0., double fRegY = 0.)
      : ALCalc01(fRegX, fRegY)
    {
      strcpy(m_cFunc, "");
    }

    void Sqrt(Logical bShowResult = TRUE,
              const char* pszMsg = "");
    void Sqr(Logical bShowResult = TRUE,
             const char* pszMsg = "");
    void Ln(Logical bShowResult = TRUE,
            const char* pszMsg = "");
    void Log(Logical bShowResult = TRUE,
             const char* pszMsg = "");
    void showFunction(const char* pszMsg = "");
    void clearRegs(); // override inherited function

  protected:
    char m_cFunc[10];
};

class ALCalc120 : virtual public ALCalc100,
                  virtual public ALCalc200
{
  public:
    ALCalc120(int nMaxMem = 10, double fRegX = 0.,
              double fRegY = 0.)
      : ALCalc100(nMaxMem, fRegX, fRegY)
    {
      strcpy(m_cFunc, "");
    }
};

//
// define member functions of class ALCalc01
//

ALCalc01::ALCalc01(double fRegX, double fRegY)
{
  m_fRegX = fRegX;
  m_fRegY = fRegY;
  m_fRegZ = 0.0;
  m_cOp = ' ';
  m_bErr = FALSE;
}

Logical ALCalc01::getError()
{
  Logical bTemp = m_bErr;
  m_bErr = FALSE;
  return bTemp;
}

void ALCalc01::add(Logical bShowResult, const char* pszMsg)
{
  m_fRegZ = m_fRegY + m_fRegX;
  m_cOp = '+';
  m_bErr = FALSE;
  if (bShowResult)
    cout << pszMsg << m_fRegZ;
}

void ALCalc01::sub(Logical bShowResult, const char* pszMsg)
{
  m_fRegZ = m_fRegY - m_fRegX;
  m_cOp = '-';
  m_bErr = FALSE;
  if (bShowResult)
    cout << pszMsg << m_fRegZ;
}

void ALCalc01::mult(Logical bShowResult, const char* pszMsg)
{
  m_fRegZ = m_fRegY * m_fRegX;
  m_cOp = '*';
  m_bErr = FALSE;
  if (bShowResult)
    cout << pszMsg << m_fRegZ;
}

void ALCalc01::div(Logical bShowResult, const char* pszMsg)
{
  if (fabs(m_fRegY) > EPSILON) {
    m_cOp = '*';
    m_bErr = FALSE;
    m_fRegZ = m_fRegY / m_fRegX;
    if (bShowResult)
      cout << pszMsg << m_fRegZ;
  }
  else {
    m_fRegZ = BAD_RESULT;
    m_cOp = ' ';
    m_bErr = TRUE;
  }
}

void ALCalc01::showRegZ(const char* pszMsg)
{
  cout << pszMsg << m_fRegZ;
}

void ALCalc01::showOperation(const char* pszMsg)
{
  cout << pszMsg << m_fRegY << " " << m_cOp << " "
       << m_fRegX << " = " << m_fRegZ << "\n";
}

void ALCalc01::clearRegs()
{
  m_fRegX = 0.0;
  m_fRegY = 0.0;
  m_fRegZ = 0.0;
  m_cOp = ' ';
  m_bErr = FALSE;
}

//
// define member functions of class ALCalc100
//

ALCalc100::ALCalc100(int nMaxMem,
              double fRegX, double fRegY)
    : ALCalc01(fRegX, fRegY)
{
  m_pfMem = new double[m_nMaxMem = nMaxMem];
  for (int i = 0; i < m_nMaxMem; i++)
    m_pfMem[i] = 0.;
}

ALCalc100::~ALCalc100()
{
  delete [] m_pfMem;
}

void ALCalc100::stoMem(int nIndex, double x)
{
  if (nIndex >= 0 && nIndex < m_nMaxMem) {
    m_pfMem[nIndex] = x;
    m_bErr = FALSE;
  }
  else
    m_bErr = TRUE;
}

void ALCalc100::rclMem(int nIndex, double& x)
{
  if (nIndex >= 0 && nIndex < m_nMaxMem) {
    x = m_pfMem[nIndex];
    m_bErr = FALSE;
  }
  else
    m_bErr = TRUE;
}

void ALCalc100::clearMem()
{
  for (int i = 0; i < m_nMaxMem; i++)
    m_pfMem[i] = 0.;
}

//
// define member functions of class ALCalc200
//

void ALCalc200::Sqrt(Logical bShowResult, const char* pszMsg)
{  if (m_fRegX >= 0.) {
    m_fRegZ = sqrt(m_fRegX);
    m_bErr = FALSE;
    strcpy(m_cFunc, "Sqrt(");
    if (bShowResult)
      cout << pszMsg << m_fRegZ;
  }
  else {
    m_fRegZ = BAD_RESULT;
    m_bErr = TRUE;
    strcpy(m_cFunc, "");
  }
}

void ALCalc200::Sqr(Logical bShowResult, const char* pszMsg)
{
  m_fRegZ = m_fRegX * m_fRegX;
  m_bErr = FALSE;
  strcpy(m_cFunc, "Sqr(");
  if (bShowResult)
    cout << pszMsg << m_fRegZ;
}

void ALCalc200::Ln(Logical bShowResult, const char* pszMsg)
{
  if (m_fRegX > 0.) {
    m_fRegZ = log(m_fRegX);
    m_bErr = FALSE;
    strcpy(m_cFunc, "Ln(");
    if (bShowResult)
      cout << pszMsg << m_fRegZ;
  }
  else {
    m_fRegZ = BAD_RESULT;
    m_bErr = TRUE;
    strcpy(m_cFunc, "");
  }
}

void ALCalc200::Log(Logical bShowResult, const char* pszMsg)
{
  if (m_fRegX > 0.) {
    m_fRegZ = log10(m_fRegX);
    m_bErr = FALSE;
    strcpy(m_cFunc, "Log10(");
    if (bShowResult)
      cout << pszMsg << m_fRegZ;
  }
  else {
    m_fRegZ = BAD_RESULT;
    m_bErr = TRUE;
    strcpy(m_cFunc, "");
  }
}

void ALCalc200::showFunction(const char* pszMsg)
{
  cout << pszMsg << m_cFunc << m_fRegX
       << ") = " << m_fRegZ << "\n";
}

void ALCalc200::clearRegs()
{
  ALCalc01::clearRegs();
  strcpy(m_cFunc, "");
}

//
// test classes
//

main()
{
  {
    ALCalc01 calculator;

    cout << "Testing class ALCalc01\n";

    // test addition
    calculator.setRegX(13.0);
    calculator.setRegY(55.0);
    calculator.add(FALSE);
    calculator.showOperation();

    // test subtraction
    calculator.setRegX(55.0);
    calculator.setRegY(123.0);
    calculator.sub(FALSE);
    calculator.showOperation();

    // test multiplication
    calculator.setRegX(13.0);
    calculator.setRegY(0.55);
    calculator.mult(FALSE);
    calculator.showOperation();

    // test division
    calculator.setRegX(113.0);
    calculator.setRegY(355.0);
    calculator.div(FALSE);
    calculator.showOperation();
  }
  cout << "\n";
  {
    ALCalc100 calculator;
    double fX;

    cout << "Testing class ALCalc100\n";

    // test addition
    calculator.setRegX(13.0);
    calculator.setRegY(55.0);
    calculator.add(FALSE);
    calculator.showOperation();
    fX = calculator.getRegZ();
    calculator.stoMem(0, fX);

    // test subtraction
    calculator.setRegX(55.0);
    calculator.setRegY(123.0);
    calculator.sub(FALSE);
    calculator.showOperation();
    fX = calculator.getRegZ();
    calculator.stoMem(1, fX);

    // test multiplication
    calculator.rclMem(0, fX);
    calculator.setRegX(fX);
    calculator.rclMem(1, fX);
    calculator.setRegY(fX);
    calculator.mult(FALSE);
    calculator.showOperation();
  }
  cout << "\n";
  {
    ALCalc120 calculator;
    double fX;

    cout << "Testing class ALCalc120\n";

    // test addition
    calculator.setRegX(13.0);
    calculator.setRegY(55.0);
    calculator.add(FALSE);
    calculator.showOperation();
    fX = calculator.getRegZ();
    calculator.stoMem(0, fX);

    // test subtraction
    calculator.setRegX(55.0);
    calculator.setRegY(123.0);
    calculator.sub(FALSE);
    calculator.showOperation();
    fX = calculator.getRegZ();
    calculator.stoMem(1, fX);

    // test multiplication
    calculator.rclMem(0, fX);
    calculator.setRegX(fX);
    calculator.rclMem(1, fX);
    calculator.setRegY(fX);
    calculator.mult(FALSE);
    calculator.showOperation();

    // test ln(x)
    calculator.Ln(FALSE);
    calculator.showFunction();

    // test square function
    calculator.Sqr(FALSE);
    calculator.showFunction();
  }
  cout << "\n";
  {
    ALCalc200 calculator;

    cout << "Testing class ALCalc200\n";

    // test addition
    calculator.setRegX(13.0);

    // test ln(x)
    calculator.Ln(FALSE);
    calculator.showFunction();

    // test square function
    calculator.Sqr(FALSE);
    calculator.showFunction();
  }

  return 0;
}

