/*

   A C++ program that illustrates the following
   small class hierarchy, which uses containment:

      Distance      Random
         |            |
         |            |
         |            |
         data members of
         |            |
         \------------/
                |
                |
             Rescue

*/

#include <iostream.h>
#include <math.h>
#include <string.h>

const double PI = 4 * atan(1);
const double RAD2DEG = 180.0 / PI;
const double INIT_SEED = 113;
const double MIN_XY = 1.0;
const double MAX_XY = 1000.0;
const double CRITIC_DIST = 10.0;
const double EPSILON = 1.0e-8;
const double INFINITY = 1.0e+50;

enum Logical { FALSE, TRUE };

double sqr(double x)
{
  return x * x;
}

double cube(double x)
{
  return x * x * x;
}

double frac(double x)
{
  return x - (long)x;
}

class Random
{
  public:
    friend class Rescue;

    Random(double fSeed = INIT_SEED)
      { m_fSeed = fSeed; }

    double getRandom();

  protected:
    double m_fSeed;
};

class Distance
{
  public:
    friend class Rescue;

    Distance(double fX1, double fY1,
             double fX2, double fY2);
    void setPoint1(double fX, double fY);
    void setPoint2(double fX, double fY);
    double getDeltaX()
      { return m_fX2 - m_fX1; }
    double getDeltaY()
      { return m_fY2 - m_fY1; }
    double getDistance()
      { return sqrt(sqr(m_fX2 - m_fX2) +
                    sqr(m_fY2 - m_fY1)); }
    double getAngle();

  protected:
    double m_fX1;
    double m_fX2;
    double m_fY1;
    double m_fY2;
    double m_fDeltaX;
    double m_fDeltaY;
};

class Rescue
{
  public:
    Rescue(double fMin, double fMax,
           double fCriticalDistance);
    ~Rescue();
    void initRescue();
    Logical searchMore();
    void getCoords();
    double getDistance()
      { return m_pDistance->getDistance(); }

  protected:
    double m_fMin;
    double m_fMax;
    double m_fShift;
    double m_fCriticalDistance;
    // declare pointers to contained classes
    Distance* m_pDistance;
    Random* m_pRandom;

    double calcRandCoord()
      { return m_pRandom->getRandom() * m_fMax + m_fMin; }
    double calcRandShift()
      { return (0.5 - m_pRandom->getRandom()) * m_fShift; }
    double checkCoord(double fX)
      { return (fX >= m_fMin && fX <= m_fMax) ?
                  fX : m_fMin + m_fMax / 2.; }
};

double Random::getRandom()
{
  m_fSeed = frac(cube(m_fSeed + PI));
  return m_fSeed;
}

Distance::Distance(double fX1, double fY1,
                   double fX2, double fY2)
{
  setPoint1(fX1, fY1);
  setPoint2(fX2, fY2);
}

void Distance::setPoint1(double fX, double fY)
{
  m_fX1 = fX;
  m_fY1 = fY;
}

void Distance::setPoint2(double fX, double fY)
{
  m_fX2 = fX;
  m_fY2 = fY;
}

double Distance::getAngle()
{
  double fDeltaY = m_fY2 - m_fY1;

  return (fabs(fDeltaY) > EPSILON) ?
    RAD2DEG * atan((m_fX2 - m_fX1) / fDeltaY) :
    INFINITY;
}

Rescue::Rescue(double fMin, double fMax,
               double fCriticalDistance)
{
  m_fMin = fMin;
  m_fMax = fMax;
  m_fShift = (fMax - fMin + 1) / 10.;
  m_fCriticalDistance = fCriticalDistance;
  m_pDistance = new Distance(0, 0, 0, 0);
  m_pRandom = new Random;
  initRescue();
}

Rescue::~Rescue()
{
  delete m_pDistance;
  delete m_pRandom;
}

void Rescue::initRescue()
{
  double fX = calcRandCoord();
  double fY = calcRandCoord();

  m_pDistance->setPoint1((m_fMin + m_fMax) / 2,
                         (m_fMin + m_fMax) / 2);
  m_pDistance->setPoint2(fX, fY);
}

Logical Rescue::searchMore()
{
  if (fabs(m_pDistance->m_fDeltaX) < 1.0 ||
      fabs(m_pDistance->m_fDeltaY) < 1.0 ||
       m_pDistance->getDistance() < m_fCriticalDistance)
    return FALSE;
  else
    return TRUE;
}

void Rescue::getCoords()
{

  cout << "Current distance = "
       << m_pDistance->getDistance() << "\n"
       << "Current angle = "
       << m_pDistance->getAngle() << " degrees\n";

  // calculate the random movement of lost person
  m_pDistance->m_fDeltaX = calcRandShift();
  m_pDistance->m_fDeltaY = calcRandShift();
  m_pDistance->setPoint2(
      checkCoord(m_pDistance->m_fX2 +
                 m_pDistance->m_fDeltaX),
      checkCoord(m_pDistance->m_fY2 +
                 m_pDistance->m_fDeltaY)
      );

  // prompt for movement of rescuer
  cout << "Enter shift in X coordinates : ";
  cin >> m_pDistance->m_fDeltaX;
  if (m_pDistance->m_fDeltaX > m_fShift) // verify limit
    m_pDistance->m_fDeltaX = m_fShift;
  cout << "Enter shift in Y coordinates : ";
  cin >> m_pDistance->m_fDeltaY;
  if (m_pDistance->m_fDeltaY > m_fShift) // verify limit
    m_pDistance->m_fDeltaY = m_fShift;
  m_pDistance->setPoint1(
       checkCoord(m_pDistance->m_fX1 +
                  m_pDistance->m_fDeltaX),
       checkCoord(m_pDistance->m_fY1 +
                  m_pDistance->m_fDeltaY)
       );
}

//
// test classes
//

main()
{
  Rescue R(MIN_XY, MAX_XY, CRITIC_DIST);

  R.getCoords();
  while (R.searchMore())
    R.getCoords();

  cout << "\n\n";
  if (R.getDistance() <= CRITIC_DIST)
    cout << "Congratulations for the rescue!";
  else
    cout << "Sorry you gave up!";

  return 0;
}

